
#include "CSTree.h"
#include "queue.h"
#include "stack.h"
#include "cat_rtree.h"
#include "set.h"

STNode *STNode_new()
{
	STNode *returnValue;

	returnValue = (STNode *) malloc(sizeof(STNode));
	memset(returnValue, 0, sizeof(STNode));

	returnValue->P_head = set_new();
	returnValue->P_tail = set_new();
	returnValue->P_tail2 = set_new();
	returnValue->S = set_new();
	returnValue->totalS = set_new();

	returnValue->processingClass = set_new();
	returnValue->nodeOrdering = set_new();

	return returnValue;
}

void STNode_free(STNode *aSTNode)
{
	set_free(aSTNode->P_head);
	set_free(aSTNode->P_tail);
	set_free(aSTNode->P_tail2);
	set_free(aSTNode->S);
	set_free(aSTNode->totalS);

	set_free(aSTNode->processingClass);
	set_free(aSTNode->nodeOrdering);

	free(aSTNode->P_head);
	free(aSTNode->P_tail);
	free(aSTNode->P_tail2);
	free(aSTNode->S);
	free(aSTNode->totalS);

	free(aSTNode->processingClass);
	free(aSTNode->nodeOrdering);
	//set_free(aSTNode->nextedTree);

	if (aSTNode->nestedTree != NULL)
	{
		STNode_freeAll(aSTNode->nestedTree);
	}
}

void STNode_freeAll(STNode *aSTNode)
{
	int noOfChild;
	STNode **child;

	int i;

	noOfChild = aSTNode->noOfChild;
	child = aSTNode->child;

	for (i = 0; i < noOfChild; i++)
	{
		STNode_freeAll(child[i]);

		STNode_free(child[i]);
		free(child[i]);
	}

	STNode_free(aSTNode);
}

void STNode_printNode(FILE *filePtr, STNode *aSTNode, char *prefix)
{
	char *newPrefix;

	fprintf(filePtr, "%s  ~~~~~~~~~~~~ Node ~~~~~~~~~~~~~~~~\n", prefix);
	fprintf(filePtr, "%sclass no.       : %d\n", prefix, aSTNode->classNo);
	fprintf(filePtr, "%sdim no.         : %d\n", prefix, aSTNode->dimNo);

	// P_head
	fprintf(filePtr, "%sP_head          : ", prefix);
	set_print(filePtr, aSTNode->P_head);
	fprintf(filePtr, "\n");

	// P_tail
	fprintf(filePtr, "%sP_tail          : ", prefix);
	set_print(filePtr, aSTNode->P_tail);
	fprintf(filePtr, "\n");

	// S
	fprintf(filePtr, "%sS               : ", prefix);
	set_print(filePtr, aSTNode->S);
	fprintf(filePtr, "\n");

	// totalS
	fprintf(filePtr, "%stotalS          : ", prefix);
	set_print(filePtr, aSTNode->totalS);
	fprintf(filePtr, "\n");
	
	// processingClass
	fprintf(filePtr, "%sprocessingClass : ", prefix);
	set_print(filePtr, aSTNode->processingClass);
	fprintf(filePtr, "\n");

	// node ordering
	fprintf(filePtr, "%snodeOrdering    : ", prefix);
	set_print(filePtr, aSTNode->nodeOrdering);
	fprintf(filePtr, "\n");

	// no. of child
	fprintf(filePtr, "%sno. of Child    : %d\n", prefix, aSTNode->noOfChild);

	// print his nested root
	if (aSTNode->nestedTree != NULL)
	{
		fprintf(filePtr, "%snested root     : \n", prefix);
		newPrefix = (char *) malloc(sizeof(char)*1024);
		strcpy(newPrefix, prefix);
		sprintf(newPrefix, "%s     ", prefix);
		STNode_printTree(filePtr, aSTNode->nestedTree, newPrefix);

		free(newPrefix);
	}
}

void STNode_printTree(FILE *filePtr, STNode *aSTNode, char *prefix)
{
	int i;
	STNode **child;
	// print itself
	STNode_printNode(filePtr, aSTNode, prefix);

	child = aSTNode->child;

	for (i = 0; i < aSTNode->noOfChild; i++)
	{
		STNode_printTree(filePtr, child[i], prefix);
	}
}

// -1 : the class no of the root node
STNode *STNode_createRoot(config *aConfig, int dimNo, mySet *remainD, mySet *globalSkyline)
{
	STNode *returnValue;

	int i;

	mySet *P_head;
	mySet *P_tail;
	mySet *S;
	mySet *totalS;

	int noOfClass;

	int noOfNumAttr;

	mySet *processingClass;
	//mySet *nodeOrdering;

	returnValue = STNode_new();

	returnValue->dimNo = dimNo;
	returnValue->classNo = -1;

	noOfNumAttr = aConfig->noOfDim-aConfig->noOfCat;
	noOfClass = aConfig->noOfClass[dimNo-noOfNumAttr];

	P_head = returnValue->P_head;
	P_tail = returnValue->P_tail;
	S = returnValue->S;
	totalS = returnValue->totalS;

	processingClass = returnValue->processingClass;

	set_copy(P_tail, remainD);
	set_copy(S, globalSkyline);
	set_copy(totalS, globalSkyline);
	
	// temp (has not yet introduced for template)
	for (i = 0; i < noOfClass; i++)
	{
		set_insert(processingClass, i);
	}

	returnValue->nestedTree = NULL;

	return returnValue;
}

//classID = getClass(dataID, dimNo);
int STNode_getClass(config *aConfig, int dataID, int dimNo)
{
	int returnValue;
	returnValue = aConfig->data[dataID].content[dimNo];

	return returnValue;
}

//classIndexSet = getClassIndex(aChild->P_tail, aChild->classNo);
mySet *STNode_getClassIndex(config *aConfig, mySet *aSet, int classNo, int dimNo)
{
	mySet *returnValue;
	int noOfElt;
	int *elt;

	int dataID;

	int classID;

	int i;

	noOfElt = aSet->noOfElt;
	elt = aSet->elt;

	returnValue = set_new();

	for (i = 0; i < noOfElt; i++)
	{
		dataID = elt[i];

		classID = STNode_getClass(aConfig, dataID, dimNo);
		if (classID == classNo)
		{
			// match
			set_insert(returnValue, i);
		}
	}

	return returnValue;
}

mySet *CSNode_findSkylineNonSkyline(config *aConfig, STNode *aChild, mySet *parentTotalS,  int dimNo, rtreeGroup *anRtreeGroup, myOrderSet *anOrderSet)
{
	mySet *returnValue;
	mySet *P_head;
	int noOfElt;
	int *elt;

	int i, j;

	int noOfEltPar;
	int *eltPar;

	int isDominate;
	int eltIDPar;
	int eltIDChild;

	int noOfNumAttr;

	record *recPar;
	record *recChild;

	record *data;

	mySet *anOrdering;

	//myOrderSet *anOrderSet;

	returnValue = set_new();

	noOfNumAttr = aConfig->noOfDim-aConfig->noOfCat;

	data = aConfig->data;

	P_head = aChild->P_head;

	noOfElt = P_head->noOfElt;
	elt = P_head->elt;

	noOfEltPar = parentTotalS->noOfElt;
	eltPar = parentTotalS->elt;

	set_copy(aChild->totalS, parentTotalS);

	anOrdering = aChild->nodeOrdering;

	//if (aConfig->ALG == 1)
	//{
		// brute force
		for (i = 0; i < noOfElt; i++)
		{
			eltIDChild = elt[i];
			recChild = &(data[eltIDChild]);

			isDominate = 0;
			for (j = 0; j < noOfEltPar; j++)
			{
				eltIDPar = eltPar[j];
				recPar = &(data[eltIDPar]);

				//*** temp
				//anOrderSet = orderSet_new();
				//orderSet_insert(anOrderSet, dimNo, &anOrdering);

				isDominate = isDominate_ordering(recPar, recChild, noOfNumAttr, anOrderSet);

				//orderSet_free(anOrderSet);
				//free(anOrderSet);

				if (isDominate == 1)
				{
					break;
				}
			}

			if (isDominate == 1)
			{
				// not a skyline
				// insert eltIDChild to  returnValue
				set_insert(returnValue, eltIDChild);
			}
			else
			{
				// is a skyline
				// place to S and totalS
				set_insert(aChild->S, eltIDChild);
				set_insert(aChild->totalS, eltIDChild);
			}
		}
	//}
	//else if (aConfig->ALG == 2)
	//{
	//}
	//else
	//{
	//}

	return returnValue;
}

// using template
mySet *STNode_generateAllPossProcessingClass(config *aConfig, STNode *aSTNode, int dimNo)
{
	mySet *returnValue;

	mySet *nodeOrdering;

	int i, j;

	ordering *templ;
	int noOfBinaryOrder;

	int *classID;
	int classNo;

	int classAfter, classBefore;

	int *isValid;

	int noOfValidClass;

	mySet *processingClass;

	int noOfCandidateElt;
	int *candidateElt;

	binaryOrder *orderList;
	ordering *cloneTempl;

	int noOfClassID;
	int isInitialValid;

	int noOfNumAttr;
	int anIndex;

	returnValue = set_new();

	noOfNumAttr = aConfig->noOfDim-aConfig->noOfCat;
	anIndex = dimNo - noOfNumAttr;

	templ = aConfig->templ[anIndex];
	noOfBinaryOrder = templ->noOfBinaryOrder;
	cloneTempl = ordering_clone(templ);
	orderList = cloneTempl->orderList;

	//set_copy(returnValue, aSTNode->processingClass);
	//(*noOfClass) = returnValue->noOfElt;

	nodeOrdering = aSTNode->nodeOrdering;
	noOfClassID = nodeOrdering->noOfElt;
	classID = nodeOrdering->elt;

	processingClass = aSTNode->processingClass;
	noOfCandidateElt = processingClass->noOfElt;
	candidateElt = processingClass->elt;
	
	//isValid = (int *) malloc(sizeof(int)*noofClassID);

	// find valid
	isInitialValid = 1;
	for (i = 0; i < noOfClassID; i++)
	{
		classNo = classID[i];
		//isValid[i] = 1;
		
		for (j = 0; j < noOfBinaryOrder; j++)
		{
			classBefore = orderList[j].classBefore;
			classAfter = orderList[j].classAfter;

			if (classNo == classBefore)
			{
				orderList[j].isMarkBefore = 1;
				//isValid[i] = 1;
			}
			else if (classNo == classAfter)
			{
				orderList[j].isMarkAfter = 1;

				if (orderList[j].isMarkBefore == 0)
				{
					//isValid[i] = 0;
					isInitialValid = 0;
					break;
				}
				else
				{
					//isValid[i] = 1;
				}
			}
		}

		if (isInitialValid == 0)
		{
			break;
		}
	}

	if (isInitialValid == 1)
	{
		if (noOfCandidateElt > 0)
		{
			isValid = (int *) malloc(sizeof(int)*noOfCandidateElt);

			// find valid
			for (i = 0; i < noOfCandidateElt; i++)
			{
				classNo = candidateElt[i];
				isValid[i] = 1;

				for (j = 0; j < noOfBinaryOrder; j++)
				{
					classBefore = orderList[j].classBefore;
					classAfter = orderList[j].classAfter;

					if (classNo == classBefore)
					{
						//orderList[j].isMarkBefore = 1;
						isValid[i] = 1;
					}
					else if (classNo == classAfter)
					{
						//orderList[j].isMarkAfter = 1;

						if (orderList[j].isMarkBefore == 0)
						{
							isValid[i] = 0;
							break;
						}
						else
						{
							isValid[i] = 1;
						}
					}
				}
			}

			noOfValidClass = 0;
			for (i = 0; i < noOfCandidateElt; i++)
			{
				classNo = candidateElt[i];
				if (isValid[i] == 1)
				{
					set_insert(returnValue, classNo);
					noOfValidClass++;
				}
			}

			free(isValid);
		}
	}


	ordering_free(cloneTempl);
	free(cloneTempl);

	return returnValue;
}

void STNode_findSAndTotalS_node_InNestedTree(mySet *currS, mySet *totalS, mySet *P_tail, STNode *aSTNode)
{
	int i;
	int noOfChild;
	STNode **child;

	set_unionNonRedundant(currS, aSTNode->S);
	set_unionNonRedundant(totalS, aSTNode->S);
	set_unionNonRedundant(P_tail, aSTNode->P_tail2);

	noOfChild = aSTNode->noOfChild;
	child = aSTNode->child;

	for (i = 0; i < noOfChild; i++)
	{
		STNode_findSAndTotalS_node_InNestedTree(currS, totalS, P_tail, child[i]);
	}

}

void STNode_findSAndTotalSInNestedTree(STNode *aChild, config *aConfig, mySet *P_tail)
{
	STNode *nestedTree;

	int noOfChild;
	STNode **child;

	//int i;

	mySet *currS;
	mySet *totalS;

	nestedTree = aChild->nestedTree;

	child = aChild->child;
	noOfChild = aChild->noOfChild;

	currS = aChild->S;
	totalS = aChild->totalS;

	
	STNode_findSAndTotalS_node_InNestedTree(currS, totalS, P_tail, nestedTree);
	
}


int STNode_findProjTabIndexWithProjDimClass(projTableWhole *projTabWhole, myClassSet *projDimClass)
{
	int returnValue;

	int noOfClass;

	projTable *projTab;

	int i;

	int isContain;

	noOfClass = projTabWhole->noOfProjTable;
	projTab = projTabWhole->projTab;

	returnValue = -1;
	for (i = 0; i < noOfClass; i++)
	{
		isContain = classSet_isContainAll(projTab[i].projDimClass, projDimClass);
		if (isContain == 1)
		{
			returnValue = i;
			break;
		}
	}

	return returnValue;
}

void STNode_exp_createChild(record *data, config *aConfig, STNode *aSTNode, int dimNo, myQueue *aQueue, rtreeGroup *anRtreeGroup, myOrderSet *anOrderSet, myClassSet *projDimClass)
{
	int noOfClass;

	projTableWhole *projTabWhole;

	STNode **child;
	STNode *aChild;

	mySet *processingClass;

	mySet *classIndexSet;

	mySet *nonSkyline;

	mySet *sibNonSkyline;

	int i, j;
	int index;

	int eltNo;

	int rtreeClassIndex;

	myClassSet *childProjDimClass;

	myOrderSet *childOrderSet;

	mySet *tempNodeOrdering;

	mySet *P_tail_temp;

	projTabWhole = aConfig->projTabWhole;

	if (aSTNode->P_tail->noOfElt <= aConfig->nl)
	{
		aSTNode->noOfChild = 0;
		set_copy(aSTNode->P_tail2, aSTNode->P_tail);
		return;
	}

	// temporarilly, generate all possible children

	//noOfClass = aConfig->noOfClass[dimNo];
	
	processingClass = STNode_generateAllPossProcessingClass(aConfig, aSTNode, dimNo);
	noOfClass = processingClass->noOfElt;
	
	aSTNode->noOfChild = noOfClass;

	if (noOfClass > 0)
	{

		child = (STNode **) malloc(sizeof(STNode *)*noOfClass);
		memset(child, 0, sizeof(STNode *)*noOfClass);

		aSTNode->child = child;

		sibNonSkyline = set_new();

		for (i = 0; i < noOfClass; i++)
		{
			aChild = STNode_new();
			child[i] = aChild;

			aChild->dimNo = dimNo;
			aChild->classNo = processingClass->elt[i];
			index = i;


			// processingClass
			//set_copy(aChild->processingClass, processingClass);
			set_copy(aChild->processingClass, aSTNode->processingClass);
			set_remove(child[i]->processingClass, index);

			// nodeOrdering
			set_copy(aChild->nodeOrdering, aSTNode->nodeOrdering);
			set_insert(aChild->nodeOrdering, aChild->classNo);

			// P_tail
			set_copy(aChild->P_tail, aSTNode->P_tail);
			classIndexSet = STNode_getClassIndex(aConfig, aChild->P_tail, aChild->classNo, dimNo);
			for (j = classIndexSet->noOfElt-1; j >= 0; j--)
			{
				set_remove(aChild->P_tail, classIndexSet->elt[j]);
			}

			// P_head
			for (j = 0; j < classIndexSet->noOfElt; j++)
			{
				index = classIndexSet->elt[j];
				eltNo = aSTNode->P_tail->elt[index];
				set_insert(aChild->P_head, eltNo);
			}

			// S
			childOrderSet = orderSet_new();
			orderSet_copy(childOrderSet, anOrderSet);
			tempNodeOrdering = set_new();
			set_copy(tempNodeOrdering, aChild->nodeOrdering);
			orderSet_insertAndUpdate(childOrderSet, aChild->dimNo, &(tempNodeOrdering));
//if ((aChild->classNo == 1) && (aChild->dimNo == 3) && (anOrderSet->dimNo[0] == 2) && (anOrderSet->orderList[0]->elt[0] == 2) && (aChild->nodeOrdering->noOfElt >= 1) && (aChild->nodeOrdering->elt[0] == 0))
//{
//	printf("");
//}
//if ((aChild->classNo == 0) && (aChild->dimNo == 3)) // && (anOrderSet->dimNo[0] == 2) && (anOrderSet->orderList[0]->elt[0] == )
//{
//	printf("");

//}
			set_copy(aChild->totalS, aSTNode->totalS);

			childProjDimClass = classSet_new();
			classSet_copy(childProjDimClass, projDimClass);
			classSet_insert(childProjDimClass, dimNo, aChild->classNo);

			if (dimNo == aConfig->noOfDim-1)
			{
				if (aConfig->ALG == 1)
				{
					nonSkyline = CSNode_findSkylineNonSkyline(aConfig, aChild, aSTNode->totalS, dimNo, anRtreeGroup, childOrderSet);
				}
				else 
				{
					if (aConfig->ALG == 2)
					{
						rtreeClassIndex = 0;
					}
					else
					{
						//rtreeClassIndex = aChild->classNo;
						rtreeClassIndex = STNode_findProjTabIndexWithProjDimClass(projTabWhole, childProjDimClass);
						//if (rtreeClassIndex == -1)
						//{
						//	//printf("has problem!\n");
						//}
					}

					if (rtreeClassIndex != -1)
					{
						rtree_findSkyline_node(data, aConfig, anRtreeGroup, aSTNode->totalS, aChild, rtreeClassIndex, childOrderSet, childProjDimClass);
					}
					nonSkyline = set_new();
					set_copy(nonSkyline, aChild->P_head);
					set_minus(nonSkyline, aChild->S);
				}
			}
			else
			{
				
				

				
				
				aChild->nestedTree = CSNode_BFS(aConfig->data, aConfig, dimNo+1, aChild->P_head, aChild->totalS, anRtreeGroup, childOrderSet, childProjDimClass);

				

				

				// find S and totalS in aChild (from the nested tree)
				P_tail_temp = set_new();
				STNode_findSAndTotalSInNestedTree(aChild, aConfig, P_tail_temp);

				

				set_unionNonRedundant(aSTNode->P_tail2, P_tail_temp);

				set_free(P_tail_temp);
				free(P_tail_temp);

				nonSkyline = set_new();
				set_copy(nonSkyline, aChild->P_head);
				set_minus(nonSkyline, aChild->S);
			}

			// free all?
			classSet_free(childProjDimClass);	
			free(childProjDimClass);

			orderSet_freeAll(childOrderSet);
			free(childOrderSet);

			// union nonskyline to sibNonSkyline
			set_union(sibNonSkyline, nonSkyline);

			// free nonSkyline
			set_free(nonSkyline);
			free(nonSkyline);

			set_free(classIndexSet);
			free(classIndexSet);
			//child[i]->P_head
			//child[i]->P_tail
			//child[i]->S
			//child[i]->nestTree
		}	

		// for sibNonSkyline and enqueue
		// remove sibNonSkyline from P_tail (of each child)
		for (i = 0; i < noOfClass; i++)
		{
			set_minus(child[i]->P_tail, sibNonSkyline);

			queue_enqueue(aQueue, &(child[i]));
		}


		set_free(sibNonSkyline);
		free(sibNonSkyline);
	}

	set_free(processingClass);
	free(processingClass);

//	set_empty(aSTNode->P_tail);
//	set_empty(aSTNode->totalS);
//	set_empty(aSTNode->processingClass);
//	set_empty(aSTNode->P_head);
}


STNode *CSNode_BFS(record *data, config *aConfig, int dimNo, mySet *remainD, mySet *globalSkyline, rtreeGroup *anRtreeGroup, myOrderSet *anOrderSet, myClassSet *projDimClass)
{
	STNode *root;
	STNode *currNode;

	myQueue *aQueue;

	int noOfNumAttr;

	mySet *tempSet;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;

	aQueue = queue_new();

	// create the root
	if (dimNo == noOfNumAttr)
	{
		root = STNode_createRoot(aConfig, dimNo, remainD, globalSkyline);
	}
	else
	{
		tempSet = set_new();
		root = STNode_createRoot(aConfig, dimNo, remainD, tempSet);
		set_free(tempSet);
		free(tempSet);

		set_copy(root->totalS, globalSkyline);
	}
	
	queue_enqueue(aQueue, &root);

	// bfs approach
	currNode = queue_dequeue(aQueue);
	while (currNode != NULL)
	{
		STNode_exp_createChild(data, aConfig, currNode, dimNo, aQueue, anRtreeGroup, anOrderSet, projDimClass);
		currNode = queue_dequeue(aQueue);
	}

	return root;
}

myStack *CSNode_put2Stack(STNode *root)
{
	myStack *returnValue;
	STNode *currNode;

	int i;

	myQueue *aQueue;

	STNode **child;

	currNode = root;
	aQueue = queue_new();
	returnValue = stack_new();

	//queue_enqueue(aQueue, &currNode);

	//if (currNode->noOfChild >= 1)
	if (currNode->noOfChild >= 0)
	{
		queue_enqueue(aQueue, &currNode);
		stack_push(returnValue, &currNode);
	}

	currNode = queue_dequeue(aQueue);
	while (currNode != NULL)
	{
		child = currNode->child;
		
		for (i = 0; i < currNode->noOfChild; i++)
		{
			if (currNode->noOfChild >= 1)
			{
				queue_enqueue(aQueue, &(child[i]));
				stack_push(returnValue, &(child[i]));
			}
		}

		currNode = queue_dequeue(aQueue);
	}

	queue_free(aQueue);
	free(aQueue);

	return returnValue;
}

//commonS = CSNode_getCommonSInChild(currNode);
mySet *CSNode_getCommonSInChild(STNode *aSTNode)
{
	int noOfChild;
	STNode **child;

	mySet *returnValue;

	int i;

	child = aSTNode->child;
	noOfChild = aSTNode->noOfChild;

	returnValue = set_new();

	if (noOfChild >= 1)
	{
		set_copy(returnValue, child[0]->S);
	}

	for (i = 1; i < noOfChild; i++)
	{
		// do intersection
		set_intersection(returnValue, child[i]->S);
	}

	return returnValue;
}



mySet *CSNode_getCommonPTailInChild(STNode *aSTNode)
{
	int noOfChild;
	STNode **child;

	mySet *returnValue;

	int i;

	child = aSTNode->child;
	noOfChild = aSTNode->noOfChild;

	returnValue = set_new();

	if (noOfChild >= 1)
	{
		set_copy(returnValue, child[0]->P_tail2);
	}

	for (i = 1; i < noOfChild; i++)
	{
		// do intersection
		set_intersection(returnValue, child[i]->P_tail2);
	}

	return returnValue;
}

void CSNode_removeCommonSFromChild(STNode *currNode, mySet *commonS)
{
	int i;

	int noOfChild;
	STNode **child;

	child = currNode->child;
	noOfChild = currNode->noOfChild;

	for (i = 0; i < noOfChild; i++)
	{
		set_minus(child[i]->S, commonS);
	}
}


void CSNode_removeCommonPTailFromChild(STNode *currNode, mySet *commonPTail)
{
	int i;

	int noOfChild;
	STNode **child;

	child = currNode->child;
	noOfChild = currNode->noOfChild;

	for (i = 0; i < noOfChild; i++)
	{
		set_minus(child[i]->P_tail2, commonPTail);
	}
}

void CSNode_insertCommonS2CurrNode(STNode *currNode, mySet *commonS)
{
	int i;

	int noOfElt;
	int *elt;

	int eltNo;

	noOfElt = commonS->noOfElt;
	elt = commonS->elt;

	for (i = 0; i < noOfElt; i++)
	{
		eltNo = elt[i];


		set_insert(currNode->S, eltNo);
	}
}

void CSNode_insertCommonPTail2CurrNode(STNode *currNode, mySet *commonPTail)
{
	int i;

	int noOfElt;
	int *elt;

	int eltNo;

	noOfElt = commonPTail->noOfElt;
	elt = commonPTail->elt;

	for (i = 0; i < noOfElt; i++)
	{
		eltNo = elt[i];


		set_insert(currNode->P_tail2, eltNo);
	}
}

void CSNode_removeChildIndex(STNode *currNode, int index)
{
	STNode **child;
	int noOfChild;

	STNode **newChild;

	STNode *childRemove;

	//int i;

	noOfChild = currNode->noOfChild;
	child = currNode->child;

	childRemove = child[index];

	STNode_free(childRemove);
	free(childRemove);

	if (noOfChild == 1)
	{
		free(child);

		currNode->child = NULL;
		currNode->noOfChild = 0;
	}
	else
	{
		newChild = (STNode **) malloc(sizeof(STNode *)*(noOfChild-1));

		if (index > 0)
		{
			memcpy(newChild, child, sizeof(STNode *)*index);
		}
		if (index < noOfChild-1)
		{
			memcpy(&(newChild[index]), &(child[index+1]), sizeof(STNode *)*(noOfChild-1-index));
		}

		currNode->child = newChild;
		currNode->noOfChild = noOfChild-1;
	}
}

void CSNode_removeEmptyChild(STNode *currNode)
{
	int i;

	int noOfChild;
	STNode **child;

	int *isMark;

	mySet *childS;
	mySet *childPTail;

	noOfChild = currNode->noOfChild;
	child = currNode->child;

	isMark = (int *) malloc(sizeof(int)*noOfChild);
	memset(isMark, 0, sizeof(int)*noOfChild);

	for (i = 0; i < noOfChild ; i++)
	{
		childS = child[i]->S;
		childPTail = child[i]->P_tail2;

		if ((childS->noOfElt == 0) && (childPTail->noOfElt == 0) && (child[i]->noOfChild == 0) && (child[i]->nestedTree == NULL))
		{			
			isMark[i] = 1;
		}
	}

	for (i = noOfChild-1; i >= 0; i--)
	{
		if (isMark[i] == 1)
		{
			CSNode_removeChildIndex(currNode, i);
		}
	}

	free(isMark);
}

int CSNode_isEmpty(STNode *aSTNode)
{
	int noOfChild;

	int returnValue;

	noOfChild = aSTNode->noOfChild;

	if (noOfChild == 0)
	{
		if (aSTNode->S->noOfElt == 0)
		{
			returnValue = 1;
		}
		else
		{
			returnValue = 0;
		}
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}

// compress the root
void CSNode_compress(STNode *root, int noOfDim)
{
	myStack *aStack;

	mySet *commonS;
	mySet *commonPTail;

	STNode *currNode;

	int count;

	int isEmpty;

	//STNode **child;

	// push the nodes into the stack
	aStack = CSNode_put2Stack(root);

	count = 0;
	currNode = stack_pop(aStack);
	while (currNode != NULL)
	{
		if (currNode->dimNo != noOfDim-1)
		{
			// do nested
			if (currNode->nestedTree != NULL)
			{
				CSNode_compress(currNode->nestedTree, noOfDim);

				if (currNode->classNo != -1)
				{
					// overwrite
					set_copy(currNode->S, currNode->nestedTree->S);
				}

				isEmpty = CSNode_isEmpty(currNode->nestedTree);
				if (isEmpty == 1)
				{
					STNode_freeAll(currNode->nestedTree);
					free(currNode->nestedTree);

					currNode->nestedTree = NULL;
				}
			}
		}

		
		// **** common S
		// get common S in child
		commonS = CSNode_getCommonSInChild(currNode);

		// remove commonS from S of each child
		CSNode_removeCommonSFromChild(currNode, commonS);

		// insert commonS to S of currNode
		CSNode_insertCommonS2CurrNode(currNode, commonS);

		// **** common P_tail
		// get common P_tail in child
		commonPTail = CSNode_getCommonSInChild(currNode);

		// remove commonP_tail from P_tail of each child
		CSNode_removeCommonPTailFromChild(currNode, commonPTail);

		// insert commonP_tail to P_tail of currNode
		CSNode_insertCommonPTail2CurrNode(currNode, commonPTail);

//if (count == 5)
//{
//	printf("");
//}
		// remove child if S (of child) is empty set and P_tail (of child) is empty set
		// (in the parent node)
		CSNode_removeEmptyChild(currNode);

		// pop the stack
		currNode = stack_pop(aStack);
		
		count++;
	}
}

//elt_P_sort = STNode_sortPTailByOrdering(elt_P_unsort, noOfElt_P, aConfig, anOrdering);
int *STNode_sortPTailByOrdering(int *elt_unsort, int noOfElt, int dimNo, config *aConfig, mySet *anOrdering, int noOfClassID)
{
	int *returnValue;

	int *orderID;
	int noOfOrderID;

	int i, j;

	int *tempIndex;

	int temp;

	int eltID;
	int no;

	int classID;

	int *mapping;

	//int *indexing;

	returnValue = (int *) malloc(sizeof(int)*noOfElt);
	memset(returnValue, 0, sizeof(int)*noOfElt);

	orderID = anOrdering->elt;
	noOfOrderID = anOrdering->noOfElt;

	//mapping = (int *) malloc(sizeof(int)*noOfOrderID);
	//memset(mapping, 0, sizeof(int)*noOfOrderID);
	mapping = (int *) malloc(sizeof(int)*noOfClassID);
	memset(mapping, 0, sizeof(int)*noOfClassID);
	for (i = 0; i < noOfClassID; i++)
	{
		mapping[i] = -1;
	}
	for (i = 0; i < noOfOrderID; i++)
	{
		no = orderID[i];
		mapping[no] = i;
	}

	tempIndex = (int *) malloc(sizeof(int)*noOfElt);
	memset(tempIndex, 0, sizeof(int)*noOfElt);

	for (i = 0; i < noOfElt; i++)
	{
		eltID = elt_unsort[i];
		classID = STNode_getClass(aConfig, eltID, dimNo);
		
		tempIndex[i] = mapping[classID];
	}

	temp = 0;
	for (i = 0; i < noOfOrderID; i++)
	{
		for (j = 0; j < noOfElt; j++)
		{	
			if (tempIndex[j] == i)
			{
				// insert
				eltID = elt_unsort[j];
				returnValue[temp] = eltID;
				temp++;
			}
		}
	}

	for (j = 0; j < noOfElt; j++)
	{
		if (tempIndex[j] == -1)
		{
			eltID = elt_unsort[j];
			returnValue[temp] = eltID;
			temp++;
		}
	}

	free(mapping);
	free(tempIndex);

	return returnValue;
}

void CSNode_generateFinalSkylineFromPTail(config *aConfig, mySet *totalS, mySet *totalP_tail, myOrderSet *anOrderSet)
{
	int noOfElt_S;
	int *elt_S;
	int eltNo_S;

	int noOfElt_P;
	int *elt_P_unsort;
	int *elt_P_sort;
	int eltNo_P;

	int i, j;

	int isDominate;
	int noOfNumAttr;

	record *rec_P;
	record *rec_S;

	mySet *tempSet;

	record *data;

	//int noOfNumAttr;
	int noOfClassID;

	//*** temp
	mySet *anOrdering;
	int dimNo;

	anOrdering = anOrderSet->orderList[0];
	dimNo = anOrderSet->dimNo[0];

	tempSet = set_new();

	data = aConfig->data;

	noOfElt_S = totalS->noOfElt;
	elt_S = totalS->elt;

	noOfElt_P = totalP_tail->noOfElt;
	elt_P_unsort = totalP_tail->elt;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
	noOfClassID = aConfig->noOfClass[dimNo-noOfNumAttr];

	if (noOfElt_P > 0)
	{
		elt_P_sort = STNode_sortPTailByOrdering(elt_P_unsort, noOfElt_P, dimNo, aConfig, anOrdering, noOfClassID);

		//noOfNumAttr = aConfig->noOfDim-aConfig->noOfCat;

		// brute force

		for (i = 0; i < noOfElt_P; i++)
		{
			eltNo_P = elt_P_sort[i];
			rec_P = &(data[eltNo_P]);

			isDominate = 0;
			for (j = 0; j < noOfElt_S; j++)
			{
				eltNo_S = elt_S[j];
				rec_S = &(data[eltNo_S]);

				//*** temp
				//anOrderSet = orderSet_new();
				//orderSet_insert(anOrderSet, dimNo, &anOrdering);

				isDominate = isDominate_ordering(rec_S, rec_P, noOfNumAttr, anOrderSet);

				//orderSet_free(anOrderSet);
				//free(anOrderSet);

				if (isDominate == 1)
				{
					break;
				}
			}

			if (isDominate == 1)
			{
				// not a skyline
				// do nothing
			}
			else
			{
				// is a skyline
				// place to tempS
				set_insert(tempSet, eltNo_P);
			}
		}
		free(elt_P_sort);
	}

	set_union(totalS, tempSet);

	set_free(tempSet);
	free(tempSet);

	
}

int CSNode_checkValidOrdering(config *aConfig, mySet *anOrdering, int dimNo)
{
	int returnValue;

	int noOfClassID;
	int *classID;

	ordering *templ;

	int i, j, k;

	int noOfBinaryOrder;
	ordering *cloneTempl;

	int classNo, classNo2;

	int classBefore, classAfter;

	binaryOrder *orderList;

	int noOfNumAttr;
	int anIndex;

	noOfNumAttr = aConfig->noOfDim-aConfig->noOfCat;
	anIndex = dimNo - noOfNumAttr;
	
	templ = aConfig->templ[anIndex];
	noOfBinaryOrder = templ->noOfBinaryOrder;
	cloneTempl = ordering_clone(templ);
	orderList = cloneTempl->orderList;

	// 
	noOfClassID = anOrdering->noOfElt;
	classID = anOrdering->elt;

	returnValue = 1;
	for (i = 0; i < noOfClassID; i++)
	{
		classNo = classID[i];

		for (j = 0; j < noOfBinaryOrder; j++)
		{
			classBefore = orderList[j].classBefore;
			classAfter = orderList[j].classAfter;

			if (classNo == classBefore)
			{
				orderList[j].isMarkBefore = 1;
				//returnValue = 1;
			}
			else if (classNo == classAfter)
			{
				orderList[j].isMarkAfter = 1;


				if (orderList[j].isMarkBefore == 0)
				{
					//**** ADDED
					// check whether the elt in the CLASSID is equal to classBefore
					int isExist;
					
					isExist = 0;
					for (k = i+1; k < noOfClassID; k++)
					{
						classNo2 = classID[k];
						if (classNo2 == classBefore)
						{
							isExist = 1;
							break;
						}
					}

					if (isExist == 1) 
					{
						// conflict
						returnValue = 0;
						break;
					}
				}
				else
				{
					//returnValue = 1;
				}
			}
		}

		if (returnValue == 0)
		{
			break;
		}
	}

	ordering_free(cloneTempl);
	free(cloneTempl);

	return returnValue;
}


int CSNode_checkValidOrderingWhole(config *aConfig, myClassSet *aClassSet)
{
	int returnValue;

	int i;

	int isValid;

	int dimNo;

	//int noOfNumAttr;

	mySet *anOrdering;

	//noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;

	returnValue = 1;
	for (i = 0; i < aClassSet->noOfElt; i++)
	{
		//dimNo = noOfNumAttr+i;
		dimNo = aClassSet->dimNo[i];
		anOrdering = aClassSet->list[i];

		isValid = CSNode_checkValidOrdering(aConfig, anOrdering, dimNo);

		if (isValid == 0)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

void CSNode_insertAllDescendant(STNode *aSTNode, mySet *returnValue, mySet *totalP_tail, int noOfDim, myOrderSet *anOrderSet, config *aConfig, int orderSetIndex)
{
	STNode *currNode;
	mySet *P_tail2;
	int i;

	int noOfChild;
	STNode **child;

	currNode = aSTNode;

	
	if (currNode->dimNo == noOfDim-1)
	{
		set_union(returnValue, currNode->S);

		P_tail2 = currNode->P_tail2;
		if (P_tail2->noOfElt != 0)
		{
			set_union(totalP_tail, P_tail2);
		}
	}
	else
	{
		//*** temp for anOrderSet
		if (currNode->nestedTree != NULL)
		{
			CSNode_findSkylineInCST_internal(returnValue, totalP_tail, currNode->nestedTree, aConfig, anOrderSet, orderSetIndex+1, NULL);
		}
	}

	noOfChild = currNode->noOfChild;
	child = currNode->child;

	for (i = 0; i < noOfChild; i++)
	{
		//*** temp for anOrderSet
		CSNode_insertAllDescendant(child[i], returnValue, totalP_tail, noOfDim, anOrderSet, aConfig, orderSetIndex);
	}
}

void CSNode_findSkylineInCST_main(mySet *skylineSet, mySet *totalP_tail, STNode *root, config *aConfig, myOrderSet *anOrderSet, int orderSetIndex, int **addFlag)
{
	mySet *returnValue;

	int i; //, j;
	//int noOfElt;
	//int *elt;

	int noOfChild;
	STNode **child;

	STNode *nestedRoot;

	int noOfClassID;
	int *classID;

	int isFound;

	int tempIndex;

	STNode *currNode;

	//int isValid;

	mySet *P_tail2;
	//mySet *totalP_tail;

	//int noOfNumAttr;

	//*** temp 
	mySet *anOrdering;
	int dimNo;

	int nestedOrderSetIndex;
	int noOfNumAttr;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;

	if (anOrderSet->noOfElt > orderSetIndex)
	{

		anOrdering = anOrderSet->orderList[orderSetIndex];
		dimNo = anOrderSet->dimNo[orderSetIndex];

		returnValue = skylineSet;

		noOfClassID = anOrdering->noOfElt;
		classID = anOrdering->elt;

		
	/*
		//returnValue = set_new();
		returnValue = skylineSet;
		

		noOfClassID = anOrdering->noOfElt;
		classID = anOrdering->elt;

		isValid = CSNode_checkValidOrdering(aConfig, anOrdering, dimNo);

		if (isValid == 0)
		{
			// do nothing
		}
		else
		{
	*/
			//
			//if (noOfClassID == 0)
			//{
			//	set_copy(returnValue, root->S);
			//}
			//else
			//{
			//totalP_tail = set_new();

		isFound = 1;
		currNode = root;
		//set_copy(returnValue, currNode->S);
		//if (dimNo == aConfig->noOfDim-1)
		{
			set_unionNonRedundant(returnValue, currNode->S);
		}
		
		P_tail2 = currNode->P_tail2;
		if (P_tail2->noOfElt != 0)
		{
			set_union(totalP_tail, P_tail2);
		}
		
		tempIndex = 0;
		while (isFound == 1)
		{
			noOfChild = currNode->noOfChild;
			child = currNode->child;
			
			isFound = 0;
			for (i = 0; i < noOfChild; i++)
			{
				if (tempIndex < noOfClassID) 
				{	
					if (child[i]->classNo == classID[tempIndex])
					{
						currNode = child[i];
						isFound = 1;

						if (dimNo == aConfig->noOfDim-1)
						{
							set_union(returnValue, currNode->S);
							
							P_tail2 = currNode->P_tail2;
							if (P_tail2->noOfElt != 0)
							{
								set_union(totalP_tail, P_tail2);
							}
						}
						else
						{
							// need to browse the nested tree
							nestedRoot = currNode->nestedTree;
							if (nestedRoot != NULL)
							{
								//noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
								nestedOrderSetIndex = orderSetIndex + 1;
								CSNode_findSkylineInCST_main(skylineSet, totalP_tail, nestedRoot, aConfig, anOrderSet, nestedOrderSetIndex, addFlag);
							}
						}
						
						//tempIndex++;
						//break;
					}
					//**** NEW
					else 
					{
						int relativeCatDim;
						int thisClassNo;

						relativeCatDim = child[i]->dimNo-noOfNumAttr;
						thisClassNo = child[i]->classNo;
						if (addFlag[relativeCatDim][thisClassNo] == 1)
						{
							addFlag[relativeCatDim][thisClassNo] = 0;

							if (dimNo == aConfig->noOfDim-1)
							{
								set_union(returnValue, child[i]->S);
								
								P_tail2 = child[i]->P_tail2;
								if (P_tail2->noOfElt != 0)
								{
									set_union(totalP_tail, P_tail2);
								}
							}
							else
							{
								// need to browse the nested tree
								nestedRoot = child[i]->nestedTree;
								if (nestedRoot != NULL)
								{
									//noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
									nestedOrderSetIndex = orderSetIndex + 1;
									CSNode_findSkylineInCST_main(skylineSet, totalP_tail, nestedRoot, aConfig, anOrderSet, nestedOrderSetIndex, addFlag);
								}
							}
						}
					}
				}
				else
				{
					// further to find the descendants (for partial order)
					// (for full order, this code cannot be executed automatically)
					
					// it has children
			//		currNode = child[i];
					//*** temp for anOrderSet
			//		CSNode_insertAllDescendant(currNode, returnValue, totalP_tail, aConfig->noOfDim, anOrderSet, aConfig, orderSetIndex);
					break;		
				}
			}

			tempIndex++;
		}
	}

			
		

		//set_free(totalP_tail);
		//free(totalP_tail);
		//}
/*		
	}

*/

	//return returnValue;
}


void CSNode_findSkylineInCST_main_allPoss(mySet *skylineSet, mySet *totalP_tail, STNode *root, config *aConfig, myOrderSet *anOrderSet, int orderSetIndex)
{
	mySet *returnValue;

	int i, j;
	//int noOfElt;
	//int *elt;

	int noOfChild;
	STNode **child;

	STNode *nestedRoot;

	int noOfClassID;
	int *classID;

	int isFound;

	int tempIndex;

	STNode *currNode;

	//int isValid;

	mySet *P_tail2;
	//mySet *totalP_tail;

	//int noOfNumAttr;

	//*** temp 
	mySet *anOrdering;
	int dimNo;

	int nestedOrderSetIndex;

	if (anOrderSet->noOfElt > orderSetIndex)
	{

		anOrdering = anOrderSet->orderList[orderSetIndex];
		dimNo = anOrderSet->dimNo[orderSetIndex];

		returnValue = skylineSet;

		noOfClassID = anOrdering->noOfElt;
		classID = anOrdering->elt;

		
	/*
		//returnValue = set_new();
		returnValue = skylineSet;
		

		noOfClassID = anOrdering->noOfElt;
		classID = anOrdering->elt;

		isValid = CSNode_checkValidOrdering(aConfig, anOrdering, dimNo);

		if (isValid == 0)
		{
			// do nothing
		}
		else
		{
	*/
			//
			//if (noOfClassID == 0)
			//{
			//	set_copy(returnValue, root->S);
			//}
			//else
			//{
			//totalP_tail = set_new();

		isFound = 1;
		currNode = root;
		//set_copy(returnValue, currNode->S);
		//if (dimNo == aConfig->noOfDim-1)
		{
			set_unionNonRedundant(returnValue, currNode->S);
		}
		
		P_tail2 = currNode->P_tail2;
		if (P_tail2->noOfElt != 0)
		{
			set_union(totalP_tail, P_tail2);
		}
		
		tempIndex = 0;
		//while (isFound == 1)
		//{
			noOfChild = currNode->noOfChild;
			child = currNode->child;
			
			isFound = 0;
			for (i = 0; i < noOfChild; i++)
			{
				if (tempIndex < noOfClassID) 
				{	
					for (j = 0; j < noOfClassID; j++)
					{
						if (child[i]->classNo == classID[j])
						{
							currNode = child[i];
							isFound = 1;

							if (dimNo == aConfig->noOfDim-1)
							{
								set_union(returnValue, currNode->S);
								
								P_tail2 = currNode->P_tail2;
								if (P_tail2->noOfElt != 0)
								{
									set_union(totalP_tail, P_tail2);
								}
							}
							else
							{
								// need to browse the nested tree
								nestedRoot = currNode->nestedTree;
								if (nestedRoot != NULL)
								{
									//noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
									nestedOrderSetIndex = orderSetIndex + 1;
									CSNode_findSkylineInCST_main_allPoss(skylineSet, totalP_tail, nestedRoot, aConfig, anOrderSet, nestedOrderSetIndex);
								}
							}
							
							tempIndex++;
						//	break;
						}
					}
				}
				else
				{
					// further to find the descendants (for partial order)
					// (for full order, this code cannot be executed automatically)
					
					// it has children
				//	currNode = child[i];
					//*** temp for anOrderSet
				//	CSNode_insertAllDescendant(currNode, returnValue, totalP_tail, aConfig->noOfDim, anOrderSet, aConfig, orderSetIndex);

					break;
				}
			}
		//}
	}

			
		

		//set_free(totalP_tail);
		//free(totalP_tail);
		//}
/*		
	}

*/

	//return returnValue;
}

void CSNode_findSkylineInCST_internal(mySet *skylineSet, mySet *totalP_tail, STNode *root, config *aConfig, myOrderSet *anOrderSet, int orderSetIndex, int **addFlag)
{
	//mySet *returnValue;

	//int i;
	//int noOfElt;
	//int *elt;

	//int noOfChild;
	//STNode **child;

	int noOfClassID;
	int *classID;

	//int orderSetIndex;

	//int isFound;

	//int tempIndex;

	//STNode *currNode;

	int isValid;

	//mySet *P_tail2;
	//mySet *totalP_tail;

	//*** temp 
	mySet *anOrdering;
	int dimNo;

	anOrdering = anOrderSet->orderList[orderSetIndex];
	dimNo = anOrderSet->dimNo[orderSetIndex];

	//returnValue = set_new();
	

	noOfClassID = anOrdering->noOfElt;
	classID = anOrdering->elt;

	isValid = CSNode_checkValidOrdering(aConfig, anOrdering, dimNo);

	if (isValid == 0)
	{
		// do nothing
	}
	else
	{
		//orderSetIndex = 0;
		CSNode_findSkylineInCST_main(skylineSet, totalP_tail, root, aConfig, anOrderSet, orderSetIndex, addFlag);

	}

	//return returnValue;	
}


void CSNode_findSkylineInCST_internal_allPoss(mySet *skylineSet, mySet *totalP_tail, STNode *root, config *aConfig, myOrderSet *anOrderSet, int orderSetIndex)
{
	//mySet *returnValue;

	//int i;
	//int noOfElt;
	//int *elt;

	//int noOfChild;
	//STNode **child;

	int noOfClassID;
	int *classID;

	//int orderSetIndex;

	//int isFound;

	//int tempIndex;

	//STNode *currNode;

	int isValid;

	//mySet *P_tail2;
	//mySet *totalP_tail;

	//*** temp 
	mySet *anOrdering;
	int dimNo;

	anOrdering = anOrderSet->orderList[orderSetIndex];
	dimNo = anOrderSet->dimNo[orderSetIndex];

	//returnValue = set_new();
	

	noOfClassID = anOrdering->noOfElt;
	classID = anOrdering->elt;

	isValid = CSNode_checkValidOrdering(aConfig, anOrdering, dimNo);

	if (isValid == 0)
	{
		// do nothing
	}
	else
	{
		//orderSetIndex = 0;
		CSNode_findSkylineInCST_main(skylineSet, totalP_tail, root, aConfig, anOrderSet, orderSetIndex, NULL);

	}

	//return returnValue;	
}


mySet *CSNode_findSkylineInCST(STNode *root, config *aConfig, myOrderSet *anOrderSet)
{
	mySet *returnValue;

	int i;
	//int noOfElt;
	//int *elt;

	//int noOfChild;
	//STNode **child;

	//int noOfClassID;
	//int *classID;

	//int orderSetIndex;

	//int isFound;

	//int tempIndex;

	//STNode *currNode;

	//int isValid;

	//mySet *P_tail2;
	mySet *totalP_tail;

	int orderSetIndex;

	int **addFlag;

	//*** temp 
	//mySet *anOrdering;
	//int dimNo;

	//anOrdering = anOrderSet->orderList[0];
	//dimNo = anOrderSet->dimNo[0];

	returnValue = set_new();
	totalP_tail = set_new();

	orderSetIndex = 0;

	addFlag = (int **) malloc(sizeof(int *)*aConfig->noOfCat);
	for (i = 0; i < aConfig->noOfCat; i++)
	{
		addFlag[i] = (int *) malloc(sizeof(int)*aConfig->noOfClass[i]);
		memset(addFlag[i], 0, sizeof(int)*aConfig->noOfClass[i]);
	}

	CSNode_findSkylineInCST_internal(returnValue, totalP_tail, root, aConfig, anOrderSet, orderSetIndex, addFlag);

	CSNode_generateFinalSkylineFromPTail(aConfig, returnValue, totalP_tail, anOrderSet);

	set_free(totalP_tail);
	free(totalP_tail);

	for (i = 0; i < aConfig->noOfCat; i++)
	{
		free(addFlag[i]);
	}
	free(addFlag);

	return returnValue;	
}


mySet *CSNode_findSkylineInCST_forPartial(STNode *root, config *aConfig, myOrderSet *anOrderSet, int **addFlag)
{
	mySet *returnValue;

	//int i;
	//int noOfElt;
	//int *elt;

	//int noOfChild;
	//STNode **child;

	//int noOfClassID;
	//int *classID;

	//int orderSetIndex;

	//int isFound;

	//int tempIndex;

	//STNode *currNode;

	//int isValid;

	//mySet *P_tail2;
	mySet *totalP_tail;

	int orderSetIndex;

	//*** temp 
	//mySet *anOrdering;
	//int dimNo;

	//anOrdering = anOrderSet->orderList[0];
	//dimNo = anOrderSet->dimNo[0];

	returnValue = set_new();
	totalP_tail = set_new();

	orderSetIndex = 0;

	CSNode_findSkylineInCST_internal(returnValue, totalP_tail, root, aConfig, anOrderSet, orderSetIndex, addFlag);

	CSNode_generateFinalSkylineFromPTail(aConfig, returnValue, totalP_tail, anOrderSet);

	set_free(totalP_tail);
	free(totalP_tail);

	return returnValue;	
}


mySet *CSNode_findSkylineInCST_allPoss(STNode *root, config *aConfig, myOrderSet *anOrderSet)
{
	mySet *returnValue;

	//int i;
	//int noOfElt;
	//int *elt;

	//int noOfChild;
	//STNode **child;

	//int noOfClassID;
	//int *classID;

	//int orderSetIndex;

	//int isFound;

	//int tempIndex;

	//STNode *currNode;

	//int isValid;

	//mySet *P_tail2;
	mySet *totalP_tail;

	int orderSetIndex;

	//*** temp 
	//mySet *anOrdering;
	//int dimNo;

	//anOrdering = anOrderSet->orderList[0];
	//dimNo = anOrderSet->dimNo[0];

	returnValue = set_new();
	totalP_tail = set_new();

	orderSetIndex = 0;

	CSNode_findSkylineInCST_internal_allPoss(returnValue, totalP_tail, root, aConfig, anOrderSet, orderSetIndex);

	CSNode_generateFinalSkylineFromPTail(aConfig, returnValue, totalP_tail, anOrderSet);

	set_free(totalP_tail);
	free(totalP_tail);

	return returnValue;	
}

void CSNode_analyzeNodeSize(STNode *aSTNode, int *noOfNode, int *noOfDataPointInS, int *noOfDataPointInPTail)
{
	(*noOfNode)++;
	(*noOfDataPointInS) += aSTNode->S->noOfElt;
	(*noOfDataPointInPTail) += aSTNode->P_tail2->noOfElt;
}

void CSNode_analyzeSize(STNode *root, int *noOfNode, int *noOfDataPointInS, int *noOfDataPointInPTail, int levelNo, int noOfCat)
{
	int noOfChild;
	STNode **child;

	int i;

	if (levelNo == 0)
	{
		if (root != NULL)
		{
			// for root
			CSNode_analyzeNodeSize(root, noOfNode, noOfDataPointInS, noOfDataPointInPTail);

			child = root->child;
			noOfChild = root->noOfChild;

			for (i = 0; i < noOfChild; i++)
			{
				CSNode_analyzeSize(child[i], noOfNode, noOfDataPointInS, noOfDataPointInPTail, levelNo, noOfCat);
			}
		}
	}
	else
	{
		if (root != NULL)
		{
			CSNode_analyzeSize(root->nestedTree, noOfNode, noOfDataPointInS, noOfDataPointInPTail, levelNo-1, noOfCat);

			child = root->child;
			noOfChild = root->noOfChild;

			for (i = 0; i < noOfChild; i++)
			{
				CSNode_analyzeSize(child[i], noOfNode, noOfDataPointInS, noOfDataPointInPTail, levelNo, noOfCat);
			}
		}
	}
}

void CSNode_findNodeWithPointID_recursive(config *aConfig, int pointID, STNode *aSTNode, myQueue *aQueue, int dimNo, myOrderSet *anOrderSet)
{
	int noOfChild;
	STNode **child;
	STNode *childNode;

	int i;
	//int dimNo;

	int isContain;

	int isFree;

	mySet *nodeOrdering;

	myOrderSet *newOrderSet;

	newOrderSet = orderSet_new();
	orderSet_copy(newOrderSet, anOrderSet);

	nodeOrdering = set_new();
	set_copy(nodeOrdering, aSTNode->nodeOrdering);
	//dimNo = aSTNode->dimNo;
	orderSet_insertAndUpdate(newOrderSet, dimNo, &nodeOrdering);

	isFree = 0;
	if (dimNo == aConfig->noOfDim-1)
	{

		isContain = set_isContain(aSTNode->S, pointID);
		if (isContain == 1)
		{
			aSTNode->anOrderSet = newOrderSet;

			queue_enqueue(aQueue, &aSTNode);

			isFree = 0;

			// no need to traverse the children
		}
		else
		{
			// need to traverse the children
			noOfChild = aSTNode->noOfChild;
			child = aSTNode->child;

			for (i = 0; i < noOfChild; i++)
			{
				childNode = child[i];
				CSNode_findNodeWithPointID_recursive(aConfig, pointID, childNode, aQueue, dimNo, newOrderSet);
			}

			isFree = 1;

		}
	}
	else
	{
		if (aSTNode->nestedTree != NULL)
		{
			CSNode_findNodeWithPointID_recursive(aConfig, pointID, aSTNode->nestedTree, aQueue, dimNo+1, newOrderSet);
		}

		// need to traverse the children
		noOfChild = aSTNode->noOfChild;
		child = aSTNode->child;

		for (i = 0; i < noOfChild; i++)
		{
			childNode = child[i];
			CSNode_findNodeWithPointID_recursive(aConfig, pointID, childNode, aQueue, dimNo, newOrderSet);
		}


		isFree = 1;
	}

	if (isFree == 1)
	{
		//orderSet_free(newOrderSet);
		free(newOrderSet);
	}
}

myQueue *CSNode_findNodeWithPointID(config *aConfig, int pointID, STNode *aSTNode)
{
	myQueue *returnValue;

	int startDimNo;
	int noOfNumAttr;

	myOrderSet *anOrderSet;

	returnValue = queue_new();

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
	//startDimNo = aConfig->noOfDim - noOfNumAttr;
	startDimNo = noOfNumAttr;

	anOrderSet = orderSet_new();

	CSNode_findNodeWithPointID_recursive(aConfig, pointID, aSTNode, returnValue, startDimNo, anOrderSet);

	orderSet_free(anOrderSet);
	free(anOrderSet);

	return returnValue;
}

mySet *CSNode_findSkylineInPartialOrder(STNode *root, config *aConfig, int dimNo, mainOrdering *aMainOrdering)
{
	mySet *returnValue;

	int i, j, k;
	int noOfElt;

	ordering *partialOrder;
	ordering **elt;

	int *index;

	mySetSet **allLinearEltList;

	myOrderSet *anOrderSet;
	myOrderSet *anOrderSet2;

	//int tempNo;

	int noOfNumAttr;
	int tempIndex;

	mySet *anOrdering;

	myClassSet *returnValueClass;

	int classN;
	//int dataID;

	//myClassSet *tempSkylineClass;


	int noOfClass;

	mySet *tempSkyline;
	mySet *tempSkyline2;
	//myClassSet *intersectClass;

	int *uniqueClassID;

	int isFirst;

	int count;
	int dimNo_temp;
	mySet *orderList_temp;

	myClassSet *aClassSet;

	myClassSetSet *allPossClass;

	mySet *anOrdering2;


	int eltNo;
	int isContain;

	int **addFlag;

	//int noOfNumAttr;

	//noOfNumAttr = aConfig->noOfDim - aConfig->noOfClass;



	//returnValue = setset_new();
	returnValue = set_new();

	noOfElt = aMainOrdering->noOfElt;
	elt = aMainOrdering->elt;

	allLinearEltList = (mySetSet **) malloc(sizeof(mySetSet *)*noOfElt);
	index = (int *) malloc(sizeof(int)*noOfElt);
	memset(index, 0, sizeof(int)*noOfElt);

	for (i = 0; i < noOfElt; i++)
	{
		partialOrder = elt[i];
		allLinearEltList[i] = ordering_generateAllLinearOrdering(partialOrder);
	}



	// create an Ordering
	isFirst = 1;
	if (noOfElt > 0)
	{
		// preprocess
		allPossClass = projTable_findAllPossClass(aConfig);
		noOfClass = allPossClass->noOfElt;

		noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;

		uniqueClassID = projTable_findUniqueIDIndex(aConfig, allPossClass);

		addFlag = (int **) malloc(sizeof(int)*aConfig->noOfCat);
		for (i = 0; i < aConfig->noOfCat; i++)
		{
			addFlag[i] = (int *) malloc(sizeof(int)*aConfig->noOfClass[i]);
			memset(addFlag[i], 0, sizeof(int)*aConfig->noOfClass[i]);
		}

		returnValueClass = classSet_new();
		while (1)
		{
			anOrderSet = orderSet_new();
			anOrderSet2 = orderSet_new();
			for (i = 0; i < noOfElt; i++)
			{
				dimNo = noOfNumAttr+i;
				
				tempIndex = index[i];
				anOrdering = allLinearEltList[i]->elt[tempIndex];
				orderSet_insert(anOrderSet, dimNo, &anOrdering);


				anOrdering2 = set_new();
				for (j = 0; j < aConfig->noOfClass[i]; j++)
				{
					eltNo = j;

					// eltNo is not in anOrdering
					isContain = set_isContain(anOrdering, eltNo);
					if (isContain == 0)
					{
						// CHANGED
						addFlag[i][j] = 1;

						set_insert(anOrdering2, eltNo);
					}
				}
				orderSet_insert(anOrderSet2, dimNo, &anOrdering2);
				
				//index[i] = ((index[i]+1)%allLinearEltList[i]->noOfElt);
			}

			// do sth
			tempSkyline = CSNode_findSkylineInCST_forPartial(root, aConfig, anOrderSet, addFlag);

			
			tempSkyline2 = CSNode_findSkylineInCST_forPartial(root,aConfig, anOrderSet2, addFlag);

			set_unionNonRedundant(tempSkyline, tempSkyline2);

			
			/*
			for (j = 0; j < returnValue->noOfElt; j++)
			{
				dataID = returnValue->elt[j];
				//classN = STNode_getClass(aConfig, dataID, dimNo);
				classN = uniqueClassID[dataID];
				set_insert(returnValueClass, classN);
			}
			*/

			if (isFirst == 1)
			{
				set_copy(returnValue, tempSkyline);
				isFirst = 0;
			}
			else
			{
				//noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
				//noOfClass = aConfig->noOfClass[dimNo-noOfNumAttr];

				set_intersection(returnValue, tempSkyline);
/*
				
				tempSkylineClass = classSet_new();
				for (j = 0; j < tempSkyline->noOfElt; j++)
				{
					dataID = tempSkyline->elt[j];
					//classN = STNode_getClass(aConfig, dataID, dimNo);
					//classN = uniqueClassID[dataID];
					for (k = 0; k < aConfig->noOfCat; k++)
					{
						dimNo_temp = noOfNumAttr+k;
						classN = aConfig->data[dataID].content[dimNo_temp];
						classSet_insert(tempSkylineClass, dimNo_temp, classN);
					}
				}

				//
				intersectClass = classSet_new();
				classSet_copy(intersectClass, returnValueClass);
				classSet_intersection(intersectClass, tempSkylineClass);

				//
				//set_intersection_class(returnValue, tempSkyline, returnValueClass, tempSkylineClass, noOfClass, intersectClass);
				set_intersection_classSet(aConfig, returnValue, tempSkyline, returnValueClass, tempSkylineClass, intersectClass);

				classSet_free(intersectClass);
				free(intersectClass);

				classSet_free(tempSkylineClass);
				
				free(tempSkylineClass);
*/

			}

			//
			// find class
			for (j = 0; j < anOrderSet->noOfElt; j++)
			{
				dimNo_temp = anOrderSet->dimNo[j];
				orderList_temp = anOrderSet->orderList[j];

				aClassSet = classSet_new();
				for (k = 0; k < orderList_temp->noOfElt ; k++)
				{
					classN = orderList_temp->elt[k];
					//classSet_insertNonRedundant(returnValueClass, dimNo_temp, classN);
					classSet_insert(returnValueClass, dimNo_temp, classN);
				}
			}

			// free skyline
			set_free(tempSkyline);
			free(tempSkyline);

			orderSet_free(anOrderSet);
			free(anOrderSet);

			// move index
			count = 1;
			index[noOfElt-count]++;
			while ((noOfElt-count >= 0) &&  (index[noOfElt-count] == allLinearEltList[noOfElt-count]->noOfElt))
			{
				index[noOfElt-count] = 0;

				count++;
				if (noOfElt-count >= 0)
				{
					index[noOfElt-count]++;
				}
				else
				{
					// do nothing
				}
			}

			if (noOfElt-count < 0)
			{
				break;
			}
		}

		classSet_free(returnValueClass);
		free(returnValueClass);	

		free(uniqueClassID);

		classSetSet_free(allPossClass);
		free(allPossClass);
	}


	// free
	for (i = 0; i < noOfElt; i++)
	{
		setset_free(allLinearEltList[i]);
		free(allLinearEltList[i]);
	}
	free(allLinearEltList);
	
	free(index);


/*

	mySet *returnValue;
	mySetSet *allLinearElt;

	int noOfNumAttr;

	mySet *returnValueClass;
	mySet *tempSkylineClass;

	int i, j, k;

	int classN;
	int dataID;

	mySet **linearEltList;

	int noOfClass;
	mySet *tempSkyline;

	mySet *intersectClass;

	myOrderSet *anOrderSet;

	returnValue = set_new();




	allLinearElt = ordering_generateAllLinearOrdering(partialOrder);

	linearEltList = allLinearElt->elt;

	for (i = 0; i < allLinearElt->noOfElt; i++)
	{
		*** temp
		anOrderSet = orderSet_new();
		orderSet_insert(anOrderSet, dimNo, &(linearEltList[i]));
		tempSkyline = CSNode_findSkylineInCST(root, aConfig, anOrderSet);

		orderSet_free(anOrderSet);
		free(anOrderSet);

		if (i == 0)
		{
			set_copy(returnValue, tempSkyline);
		}
		else
		{
			noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
			noOfClass = aConfig->noOfClass[dimNo-noOfNumAttr];

			// find class
			returnValueClass = set_new();
			for (j = 0; j < returnValue->noOfElt; j++)
			{
				dataID = returnValue->elt[j];
				classN = STNode_getClass(aConfig, dataID, dimNo);
				set_insert(returnValueClass, classN);
			}
			tempSkylineClass = set_new();
			for (j = 0; j < tempSkyline->noOfElt; j++)
			{
				dataID = tempSkyline->elt[j];
				classN = STNode_getClass(aConfig, dataID, dimNo);
				set_insert(tempSkylineClass, classN);
			}

			//
			intersectClass = set_new();
			set_copy(intersectClass, returnValueClass);
			set_intersection(intersectClass, tempSkylineClass);

			//
			set_intersection_class(returnValue, tempSkyline, returnValueClass, tempSkylineClass, noOfClass, intersectClass);

			set_free(intersectClass);
			free(intersectClass);

			set_free(returnValueClass);
			set_free(tempSkylineClass);
			free(returnValueClass);
			free(tempSkylineClass);
		}

		// free skyline
		set_free(tempSkyline);
		free(tempSkyline);
	}
*/
	return returnValue;
}

