
#include "cat_rtree.h"
#include "config.h"
#include "CSTree.h"
#include "heap.h"
#include "global.h"

void rtree_start(config *aConfig)
{
	m = aConfig->m; //l_m;
	M = aConfig->M; //l_M;
	reinsert_p = aConfig->reinsert_p; //l_reinsert_p;

	dim = aConfig->noOfDim-aConfig->noOfCat;
}



//floatData = rtree_generateFloatRec(aConfig, eltNo);
float *rtree_generateFloatRec(record *data, config *aConfig, int eltNo, int noOfNumAttr)
{
	float *returnValue;
	//record *data;

	record *aRecord;

	int i;

	returnValue = (float *) malloc(sizeof(float)*noOfNumAttr);
	memset(returnValue, 0, sizeof(float)*noOfNumAttr);

	//data = aConfig->data;
	aRecord = &(data[eltNo]);

	for (i = 0; i < noOfNumAttr; i++)
	{
		returnValue[i] = (float) aRecord->content[i];
	}

	return returnValue;
}

rtreeGroup *rtree_buildMultipleTreeAndFindGlobalSkyline(config *aConfig, projTableWhole *projTabWhole, mySet *remainD, mySet *globalSkyline)
{
	rtreeGroup *returnValue;

	//int dimNo;

	int i, j;
	
	int noOfProjTab;
	projTable *projTab;

	int noOfData;
	record *data;

	int *isMark;

	int eltNo;

	int *pointID;

	int dataID;
	float *floatData;

	int noOfNumAttr;

	//int classID;

	node_type **rtreeList;

	myClassSet *projDimClass;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;


	noOfProjTab = projTabWhole->noOfProjTable;
	projTab = projTabWhole->projTab;

	// returnValue
	returnValue = (rtreeGroup *) malloc(sizeof(rtreeGroup));
	memset(returnValue, 0, sizeof(rtreeGroup));

	returnValue->noOfRtree = noOfProjTab;
	rtreeList = (node_type **) malloc(sizeof(node_type)*noOfProjTab);
	memset(rtreeList, 0, sizeof(node_type)*noOfProjTab);
	returnValue->rtreeList = rtreeList;

	for (i = 0; i < noOfProjTab; i++)
	{
		//*** should be changed
		//classID = i;
		//dimNo = aConfig->noOfDim-1;
		projDimClass = projTab[i].projDimClass;

		//
		tree_node_allocate(&(rtreeList[i]));

		for(j=0; j < dim; j++) { 
			rtreeList[i]->b[j] = (float)(-1 * INT_MAX);
			rtreeList[i]->a[j] = (float)(INT_MAX);
		}
		rtreeList[i]->id = -3;		//NO_ID
		rtreeList[i]->attribute=ROOT;
		rtreeList[i]->vacancy=M;   
		rtreeList[i]->parent=NULL; 
		for(j=0; j < M; j++)
			rtreeList[i]->ptr[j]=NULL;

		//
		noOfData = projTab[i].noOfData;
		data = projTab[i].data;
		isMark = projTab[i].isMark;
		pointID = projTab[i].pointID;

		for (j = 0; j < noOfData; j++)
		{
			if (isMark[j] == 1)
			{
				// global skyline
				eltNo = pointID[j];
				set_insert(globalSkyline, eltNo);
			}
			else if (isMark[j] == 2)
			{
				eltNo = pointID[j];
				set_insert(remainD, eltNo);

				floatData = rtree_generateFloatRec(aConfig->data, aConfig, eltNo, noOfNumAttr);
				dataID = eltNo;
				//insert_node(rtreeList[i], floatData, dataID, dimNo, classID);
				insert_node(rtreeList[i], floatData, dataID, projDimClass);

				

				free(floatData);
			}
		}
	}	

	return returnValue;
}


rtreeGroup *rtree_buildMultipleTreeAndFindLocalSkyline_preprocess(config *aConfig, projTableWhole *projTabWhole) 
//, mySet *remainD, mySet *globalSkyline)
{
	rtreeGroup *returnValue;

	int i, j;
	//int dimNo;
	
	int noOfProjTab;
	projTable *projTab;

	int noOfData;
	record *data;

	int *isMark;

	int eltNo;

	int *pointID;

	int dataID;
	float *floatData;

	int noOfNumAttr;

	//int classID;

	node_type **rtreeList;

	myClassSet *projDimClass;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;


	noOfProjTab = projTabWhole->noOfProjTable;
	projTab = projTabWhole->projTab;

	// returnValue
	returnValue = (rtreeGroup *) malloc(sizeof(rtreeGroup));
	memset(returnValue, 0, sizeof(rtreeGroup));

	returnValue->noOfRtree = noOfProjTab;
	rtreeList = (node_type **) malloc(sizeof(node_type)*noOfProjTab);
	memset(rtreeList, 0, sizeof(node_type)*noOfProjTab);
	returnValue->rtreeList = rtreeList;

	for (i = 0; i < noOfProjTab; i++)
	{
		//*** should be changed
		//classID = i;
		//dimNo = aConfig->noOfDim-1;
		projDimClass = projTab[i].projDimClass;

		//
		tree_node_allocate(&(rtreeList[i]));

		for(j=0; j < dim; j++) { 
			rtreeList[i]->b[j] = (float)(-1 * INT_MAX);
			rtreeList[i]->a[j] = (float)(INT_MAX);
		}
		rtreeList[i]->id = -3;		//NO_ID
		rtreeList[i]->attribute=ROOT;
		rtreeList[i]->vacancy=M;   
		rtreeList[i]->parent=NULL; 
		for(j=0; j < M; j++)
			rtreeList[i]->ptr[j]=NULL;

		//
		noOfData = projTab[i].noOfData;
		data = projTab[i].data;
		isMark = projTab[i].isMark;
		pointID = projTab[i].pointID;

		for (j = 0; j < noOfData; j++)
		{
			//if (isMark[j] == 1)
			//{
			//	// global skyline
			//	eltNo = pointID[j];
			//	set_insert(globalSkyline, eltNo);
			//}
			//else if (isMark[j] == 2)
			//{
				//eltNo = pointID[j];
				eltNo = j;
				//set_insert(remainD, eltNo);
if (j%1000 == 0)
{
	printf("(dimNo:%d, point:%d)\n", i, j);
}
				floatData = rtree_generateFloatRec(data, aConfig, eltNo, noOfNumAttr);
				dataID = eltNo;
				//insert_node(rtreeList[i], floatData, dataID, dimNo, classID);
//if ((i == 2) && (j == 0))
//{
//	printf("");
//}
				insert_node(rtreeList[i], floatData, dataID, projDimClass);

				free(floatData);
			//}
		}
	}	

	return returnValue;
}

rtreeGroup *rtree_buildSingleTreeAndFindGlobalSkyline(config *aConfig, projTableWhole *projTabWhole, mySet *remainD, mySet *globalSkyline)
{
	rtreeGroup *returnValue;

	int i, j;
	//int dimNo;
	
	int noOfProjTab;
	projTable *projTab;

	int noOfData;
	record *data;

	int *isMark;

	int eltNo;

	int *pointID;

	int dataID;
	float *floatData;

	int noOfNumAttr;

	//int classID;

	int fixedIndex;

	node_type **rtreeList;

	myClassSet *projDimClass;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;


	noOfProjTab = projTabWhole->noOfProjTable;
	projTab = projTabWhole->projTab;

	// returnValue
	returnValue = (rtreeGroup *) malloc(sizeof(rtreeGroup));
	memset(returnValue, 0, sizeof(rtreeGroup));

	//returnValue->noOfRtree = noOfProjTab;
	//rtreeList = (node_type **) malloc(sizeof(node_type)*noOfProjTab);
	//memset(rtreeList, 0, sizeof(node_type)*noOfProjTab);
	//returnValue->rtreeList = rtreeList;

	// single r-tree
	returnValue->noOfRtree = 1;
	rtreeList = (node_type **) malloc(sizeof(node_type)*1);
	memset(rtreeList, 0, sizeof(node_type)*1);
	returnValue->rtreeList = rtreeList;

	// 
	fixedIndex = 0;

	tree_node_allocate(&(rtreeList[fixedIndex]));

	for(j=0; j < dim; j++) { 
		rtreeList[fixedIndex]->b[j] = (float)(-1 * INT_MAX);
		rtreeList[fixedIndex]->a[j] = (float)(INT_MAX);
	}
	rtreeList[fixedIndex]->id = -3;		//NO_ID
	rtreeList[fixedIndex]->attribute=ROOT;
	rtreeList[fixedIndex]->vacancy=M;   
	rtreeList[fixedIndex]->parent=NULL; 
	for(j=0; j < M; j++)
		rtreeList[fixedIndex]->ptr[j]=NULL;

	for (i = 0; i < noOfProjTab; i++)
	{
		//*** should be changed
		//classID = i;
		//dimNo = aConfig->noOfDim-1;

		projDimClass = projTab[i].projDimClass;

		//
		//tree_node_allocate(&(rtreeList[fixedIndex]));

		//for(j=0; j < dim; j++) { 
		//	rtreeList[fixedIndex]->b[j] = (float)(-1 * INT_MAX);
		//	rtreeList[fixedIndex]->a[j] = (float)(INT_MAX);
		//}
		//rtreeList[fixedIndex]->id = -3;		//NO_ID
		//rtreeList[fixedIndex]->attribute=ROOT;
		//rtreeList[fixedIndex]->vacancy=M;   
		//rtreeList[fixedIndex]->parent=NULL; 
		//for(j=0; j < M; j++)
		//	rtreeList[fixedIndex]->ptr[j]=NULL;

		//
		noOfData = projTab[i].noOfData;
		data = projTab[i].data;
		isMark = projTab[i].isMark;
		pointID = projTab[i].pointID;

		for (j = 0; j < noOfData; j++)
		{
			if (isMark[j] == 1)
			{
				// global skyline
				eltNo = pointID[j];
				set_insert(globalSkyline, eltNo);
			}
			else if (isMark[j] == 2)
			{
				eltNo = pointID[j];
				set_insert(remainD, eltNo);

				floatData = rtree_generateFloatRec(aConfig->data, aConfig, eltNo, noOfNumAttr);
				dataID = eltNo;
				//insert_node(rtreeList[fixedIndex], floatData, dataID, dimNo, classID);
				insert_node(rtreeList[fixedIndex], floatData, dataID, projDimClass);

				free(floatData);
			}
		}
	}	

	return returnValue;
}

double rtree_calculateDist(node_type *aNode)
{
	double returnValue;

	int i;

	returnValue = 0;
	for (i = 0; i < dim; i++)
	{
		returnValue += aNode->a[i];
	}

	return returnValue;
}

//isDominate = rtree_isDominateInTotalS(aConfig, pointID, totalS);
int rtree_isDominateInTotalS_classGroup(record *data, config *aConfig, record *recordCheck, myClassSet *classGroup, mySet *totalS, myOrderSet *anOrderSet)
{
	int returnValue;
	int i;

	int noOfElt;
	int *elt;

	int isDominate;

	//record *data;

	record *recordS;

	int noOfNumAttr;

	int pointID;

	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;

	//data = aConfig->data;

	elt = totalS->elt;
	noOfElt = totalS->noOfElt;

//printf("   @@@@@@@@@@@@ 1\n");
	returnValue = 0;
	for (i = 0; i < noOfElt; i++)
	{
//printf("   @@@@@@@@@@@@ 2.%d.1\n", i);
		pointID = elt[i];
		recordS = &(data[pointID]);
//printf("   @@@@@@@@@@@@ 2.%d.2 (dimNo:%d)\n", i, dimNo);
		isDominate = isDominate_ordering_classGroup(recordS, recordCheck, classGroup, noOfNumAttr, anOrderSet);
//printf("   @@@@@@@@@@@@ 2.%d.3\n", i);
		if (isDominate == 1)
		{
			returnValue = 1;
			break;
		}
//printf("   @@@@@@@@@@@@ 2.%d.4\n", i);
	}
//printf("   @@@@@@@@@@@@ 3\n");
	return returnValue;
}

//recordCheck = generateRecord(aConfig, aNode, dimNo);
record *generateRecord(config *aConfig, node_type *aNode)
{
	record *returnValue;

	int i;

	returnValue = (record *) malloc(sizeof(record));
	returnValue->content = (int *) malloc(sizeof(int)*aConfig->noOfDim);
	memset(returnValue->content, 0, sizeof(int)*aConfig->noOfDim);

	for (i = 0; i < dim; i++)
	{
		returnValue->content[i] = (int) aNode->a[i];
	}

	//returnValue->content[dimNo] = classID;

	return returnValue;
}

void rtree_findSkyline_node(record *data, config *aConfig, rtreeGroup *anRtreeGroup, mySet *totalS, STNode *aChild, int rtreeClassIndex, myOrderSet *anOrderSet, myClassSet *projDimClass)
{
	double dist;
	node_type *aNode;

	int noOfChild;

	node_type *anRtreeNode;

	int i;

	myHeap *aHeap;

	heapEntry *aHeapEntry;
	int pointID;

	int isDominate; 

	//mySet *classGroup;
	myClassSet *classGroup;

	record *recordCheck;

	node_type *childNode;

	mySet *totalS_combine;

	int dataID;

	

	mySet *currS;
	mySet *currTotalS;
	//mySet *anOrdering;
	//int classID;

	int isContain;

	//int dimNo;

	totalS_combine = set_new();
	set_copy(totalS_combine, totalS);

	//
	currS = aChild->S;
	//*** should be changed
	//anOrdering = aChild->nodeOrdering;
	//classID = aChild->classNo;
	currTotalS = aChild->totalS;
//printf(" !!!!!!!!!!!!!! 1\n");
	// 
	set_copy(currTotalS, totalS);

	//*** should be changed
	//dimNo = anOrderSet->dimNo[0];

	aHeap = heap_new();

	anRtreeNode = anRtreeGroup->rtreeList[rtreeClassIndex];
//printf(" !!!!!!!!!!!!!! 2\n");
	// insert all entries of root node
	noOfChild = M - anRtreeNode->vacancy;
	for (i = 0; i < noOfChild; i++)
	{
		aNode = anRtreeNode->ptr[i];

		dist = rtree_calculateDist(aNode);

		heap_insertElt(aHeap, dist, &aNode);
	}
//printf(" !!!!!!!!!!!!!! 3\n");
	// while heap is non-empty
	while (aHeap->noOfElt != 0)
	{
		// remove the top entry
		aHeapEntry = heap_removeElt(aHeap);
//printf(" !!!!!!!!!!!!!! 3.1\n");
		// check the entry
		aNode = aHeapEntry->aNode;
		pointID = aNode->id;

		//isContain = set_isContain(aNode->classGroup, classID);
		//isContain = classSet_isContain(aNode->classGroup, dimNo, classID);
		isContain = classSet_isContainAll(aNode->classGroup, projDimClass);
//printf(" !!!!!!!!!!!!!! 3.2\n");
		if (isContain == 1)
		{
//printf(" !!!!!!!!!!!!!! 3.3\n");
			//
			recordCheck = generateRecord(aConfig, aNode);
			classGroup = aNode->classGroup;
//printf(" !!!!!!!!!!!!!! 3.4\n");
			// is this dominated by a point in totalS_combine
			isDominate = rtree_isDominateInTotalS_classGroup(data, aConfig, recordCheck, classGroup, totalS_combine, anOrderSet);
//printf(" !!!!!!!!!!!!!! 3.5\n");
			// free
			free(recordCheck->content);
			free(recordCheck);
//printf(" !!!!!!!!!!!!!! 3.6\n");
			// not dominated in a point in totalS_combine
			if (isDominate == 0)
			{
				// check if e is an internal node
				if (aNode->attribute != LEAF)
				{
					// internal node

					// put into the heap
					noOfChild = M - aNode->vacancy;
					for (i = 0; i < noOfChild; i++)
					{
						childNode = aNode->ptr[i];

						//
						recordCheck = generateRecord(aConfig, childNode);
						classGroup = childNode->classGroup;

						// is this dominated by a point in totalS
						isDominate = rtree_isDominateInTotalS_classGroup(data, aConfig, recordCheck, classGroup, totalS_combine, anOrderSet);

						// free
						free(recordCheck->content);
						free(recordCheck);

						dist = rtree_calculateDist(childNode);
						heap_insertElt(aHeap, dist, &(childNode));
					}
				}
				else
				{

					// leaf node
					// insert it into currS
					dataID = aNode->id;
					set_insert(currS, dataID);

					// also the totalS_combine
					set_insert(totalS_combine, dataID);

					// also the  currTotalS
					set_insert(currTotalS, dataID);
				}
			}
			else
			{
				// do nothing
			}
		}
//printf(" !!!!!!!!!!!!!! 3.end\n");
	}
//printf(" !!!!!!!!!!!!!! 4\n");
	heap_free(aHeap);
	free(aHeap);
//printf(" !!!!!!!!!!!!!! 5\n");
	set_free(totalS_combine);
	free(totalS_combine);
}



void rtree_findSkyline_preprocess(record *data, config *aConfig, rtreeGroup *anRtreeGroup, mySet *totalS, int dimNo, int rtreeClassIndex, myClassSet *projDimClass)
{
	STNode *aChild;

	int i;

	myOrderSet *anOrderSet;

	aChild = STNode_new();

	//aChild->S = ;
	//set_insert(aChild->nodeOrdering, classID);
	//aChild->classNo = classID;
	aChild->dimNo = dimNo;
	//aChild->currTotalS = ;
//printf("  *************** 2.1.1\n");

	//*** to be changed
	anOrderSet = orderSet_new();
	//orderSet_insert(anOrderSet, aChild->dimNo, &(aChild->nodeOrdering));
	for (i = 0; i < projDimClass->noOfElt; i++)
	{
		orderSet_insert(anOrderSet, projDimClass->dimNo[i], &(projDimClass->list[i]));
	}

	rtree_findSkyline_node(data, aConfig, anRtreeGroup, totalS, aChild, rtreeClassIndex, anOrderSet, projDimClass);
//printf("  *************** 2.1.2\n");
	set_copy(totalS, aChild->S);

	orderSet_free(anOrderSet);
	free(anOrderSet);
	
	STNode_free(aChild);
	free(aChild);
}

