
#include "context.h"
#include "ordering.h"
#include "config.h"
#include "CSTree.h"
#include "queue.h"

context *context_new()
{
	context *returnValue;

	returnValue = (context *) malloc(sizeof(context));
	memset(returnValue, 0, sizeof(context));

	return returnValue;
}

void context_free(context *aContext)
{
	int i;

	if (aContext->noOfAndSet > 0)
	{
		for (i = 0; i < aContext->noOfAndSet; i++)
		{
			ordering_free(&(aContext->andSet[i]));
		}
		free(aContext->andSet);
	}
}

ordering *context_generateAllBinaryOrdering(config *aConfig, int dimNo, mySet *anOrdering)
{
	ordering *returnValue;
	//mySet *anOrdering;

	int noOfClassID;
	int *classID;

	int noOfClass, noOfNumAttr;

	int i, j;

	int classElt1;
	int classElt2;

	mySet *nonClassIDSet;

	int noOfNonClassID;
	int *nonClassID;

	returnValue = ordering_new();
	
	//anOrdering = aSTNode->nodeOrdering;
	noOfClassID = anOrdering->noOfElt;
	classID = anOrdering->elt;

	// within the node ordering
	for (i = 0; i < noOfClassID; i++)
	{
		classElt1 = classID[i];
		for (j = i+1; j < noOfClassID; j++)
		{
			// generate C_i < C_j
			classElt2 = classID[j];

			ordering_insertBinaryOrder2Ordering(returnValue, classElt1, classElt2);
		}
	}

	// find noOfNonClassID and nonClassID
	nonClassIDSet = set_new();
	noOfNumAttr = aConfig->noOfDim - aConfig->noOfCat;
	noOfClass = aConfig->noOfClass[dimNo-noOfNumAttr];
	for (i = 0; i < noOfClass; i++)
	{
		set_insert(nonClassIDSet, i);
	}
	set_minus(nonClassIDSet, anOrdering);

	noOfNonClassID = nonClassIDSet->noOfElt;
	nonClassID = nonClassIDSet->elt;

	// between the node ordering and non-node ordering class
	for (i = 0; i < noOfClassID; i++)
	{
		classElt1 = classID[i];
		for (j = 0; j < noOfNonClassID; j++)
		{
			// generate C_i < C_j
			classElt2 = nonClassID[j];

			ordering_insertBinaryOrder2Ordering(returnValue, classElt1, classElt2);
		}
	}

	// free
	set_free(nonClassIDSet);
	free(nonClassIDSet);

	return returnValue;
}


mainContext *mainContext_new()
{
	mainContext *returnValue;

	returnValue = (mainContext *) malloc(sizeof(mainContext));
	memset(returnValue, 0, sizeof(mainContext));

	return returnValue;
}


void mainContext_insert(mainContext *aMainContext, context **aContext)
{
	int noOfElt;

	noOfElt = aMainContext->noOfElt;
	if (noOfElt == 0)
	{
		aMainContext->elt = (context **) malloc(sizeof(context *));
		aMainContext->elt[0] = (*aContext);
		aMainContext->noOfElt++;
	}
	else
	{
		aMainContext->elt = (context **) realloc(aMainContext->elt, sizeof(context *)*(noOfElt+1));

		aMainContext->elt[noOfElt] = (*aContext);
		aMainContext->noOfElt++;
	}
}


mainContext *context_generateContext(config *aConfig, int pointID, STNode *root)
{
	mainContext *returnValue;
	myQueue *aQueue;

	int noOfNodeList;
	STNode **nodeList;
	STNode *oneNode;

	int i, j;

	int dimNo;

	ordering *anOrdering;
	mySet *oneOrdering;

	ordering *andSet;

	myOrderSet *anOrderSet;

	//mySet *nodeOrdering;

	context *aContext;

	returnValue = mainContext_new();

	

	// find the nodes with point ID
	aQueue = CSNode_findNodeWithPointID(aConfig, pointID, root);

	noOfNodeList = aQueue->noOfElt;
	nodeList = aQueue->elt;

	

	for (i = 0; i < noOfNodeList; i++)
	{
		oneNode = nodeList[i];

		anOrderSet = oneNode->anOrderSet;


		aContext = context_new();
		aContext->noOfAndSet = anOrderSet->noOfElt;
		andSet = (ordering *) malloc(sizeof(ordering)*aContext->noOfAndSet);
		memset(andSet, 0, sizeof(ordering)*aContext->noOfAndSet);
		aContext->andSet = andSet;

		mainContext_insert(returnValue, &aContext);

//if (i == 6)
//{
//	printf("");
//}

		for (j = 0; j < anOrderSet->noOfElt; j++)
		{

			dimNo = anOrderSet->dimNo[j];
			oneOrdering = anOrderSet->orderList[j];
			//oneOrdering = oneNode->nodeOrdering;
			anOrdering = context_generateAllBinaryOrdering(aConfig, dimNo, oneOrdering);
			memcpy(&(andSet[j]), anOrdering, sizeof(ordering));

			// JUST free anOrdering (content should not be freed)
			free(anOrdering);
		}
	}

	queue_free(aQueue);
	free(aQueue);

	return returnValue;
}

void context_printContext(FILE *filePtr, context *aContext)
{
	int noOfAndSet;
	ordering *andSet;

	int i;

	noOfAndSet = aContext->noOfAndSet;
	andSet = aContext->andSet;

	for (i = 0; i < noOfAndSet; i++)
	{
		ordering_printOrdering(filePtr, &(andSet[i]));

		if (i != noOfAndSet -1 )
		{
			fprintf(filePtr, " OR ");
		}
	}
}

void mainContext_printMainContext(FILE *filePtr, mainContext *aMainContext)
{
	int noOfElt;
	context **elt;

	int i;

	noOfElt = aMainContext->noOfElt;
	elt = aMainContext->elt;

	fprintf(filePtr, "\n");
	for (i = 0; i < noOfElt; i++)
	{
		fprintf(filePtr, "  Context %d: ", i);
		context_printContext(filePtr, elt[i]);
		fprintf(filePtr, "\n");
	}
}

void context_intersection(context *aContext1, context *aContext2)
{
	int noOfElt1, noOfElt2;
	ordering *order1;
	ordering *order2;

	int i, j, k, l;
	int count;
	int noOfIntersect;
	int *isFlag;

	int isExist;
	int noOfBinaryOrder1, noOfBinaryOrder2;

	binaryOrder *orderList1;
	binaryOrder *orderList2;

	int classBefore1, classAfter1;
	int classBefore2, classAfter2;

	order1 = aContext1->andSet;
	order2 = aContext2->andSet;

	noOfElt1 = aContext1->noOfAndSet;
	noOfElt2 = aContext2->noOfAndSet;

	if (noOfElt1 > 0)
	{
		

		for (i = 0; i < noOfElt1; i++)
		{
			

			noOfBinaryOrder1 = order1[i].noOfBinaryOrder;
			orderList1 = order1[i].orderList;

			isFlag = (int *) malloc(sizeof(int)*noOfBinaryOrder1);
			memset(isFlag, 0, sizeof(int)*noOfBinaryOrder1);

			for (j = 0; j < noOfElt2; j++)
			{
				if (i == j)
				{
					noOfBinaryOrder2 = order2[j].noOfBinaryOrder;
					orderList2 = order2[j].orderList;

					for (k = 0; k < noOfBinaryOrder1; k++)
					{
						classBefore1 = orderList1[k].classBefore;
						classAfter1 = orderList1[k].classAfter;

						isExist = 0;
						for (l = 0; l < noOfBinaryOrder2; l++)
						{
							classBefore2 = orderList2[l].classBefore;
							classAfter2 = orderList2[l].classAfter;

							if ((classBefore1 == classBefore2) && (classAfter1 == classAfter2))
							{
								isExist = 1;
								break;
							}
						}

						//
						if (isExist == 1)
						{
							isFlag[k] = 1;
						}
					}
				}
			}

			noOfIntersect = 0;
			for (k = 0; k < noOfBinaryOrder1; k++)
			{
				if (isFlag[k] == 1)
				{
					noOfIntersect++;
				}
			}

			order1[i].orderList = (binaryOrder *) malloc(sizeof(binaryOrder)*noOfBinaryOrder1);
			memset(order1[i].orderList, 0, sizeof(binaryOrder)*noOfBinaryOrder1);

			count = 0;
			for (k = 0; k < noOfBinaryOrder1; k++)
			{
				if (isFlag[k] == 1)
				{
					memcpy(&(order1[i].orderList[count]), &(orderList1[k]), sizeof(binaryOrder));
					count++;
				}
			}

			order1[i].noOfBinaryOrder = noOfIntersect;

			free(isFlag);
		}

		
	}

}

context *mainContext_intersectionAll(mainContext *aMainContext)
{
	context *returnValue;
	int noOfElt;
	context **elt;

	int i;

	noOfElt = aMainContext->noOfElt;
	elt = aMainContext->elt;

	returnValue = context_new();

	if (noOfElt > 0)
	{
		memcpy(returnValue, elt[0], sizeof(context));
		for (i = 1; i < noOfElt; i++)
		{
			context_intersection(returnValue, elt[i]);
		}
	}

	return returnValue;
}

void context_negate(context *aContext)
{
	int noOfAndSet;
	ordering *andSet;

	int i, j;

	int noOfBinaryOrder;
	binaryOrder *orderList;

	int classBefore, classAfter;
	
	noOfAndSet = aContext->noOfAndSet;
	andSet = aContext->andSet;

	for (i = 0; i < noOfAndSet; i++)
	{
		noOfBinaryOrder = andSet[i].noOfBinaryOrder;
		orderList = andSet[i].orderList;

		for (j = 0; j < noOfBinaryOrder; j++)
		{
			classAfter = orderList[j].classAfter;
			classBefore = orderList[j].classBefore;

			orderList[j].classAfter = classBefore;
			orderList[j].classBefore = classAfter;
		}
	}
}

//context

//aMainContext = context_generateContext(aConfig, pointID, root);
context *context_generateDisqualify(config *aConfig, int pointID, STNode *root)
{
	mainContext *aMainContext;
	context *returnValue;

	aMainContext = context_generateContext(aConfig, pointID, root);
	//context *mainContext_intersectionAll(mainContext *aMainContext)
	returnValue = mainContext_intersectionAll(aMainContext);
	context_negate(returnValue);

	return returnValue;
}

