
#include "heap.h"

#define PRESIZE 1000

myHeap *heap_new()
{
	myHeap *returnValue;

	returnValue = (myHeap *) malloc(sizeof(myHeap));
	memset(returnValue, 0, sizeof(myHeap));

	returnValue->size = PRESIZE;

	returnValue->elt = (heapEntry *) malloc(sizeof(heapEntry)*returnValue->size);

	return returnValue;
}

void heap_free(myHeap *aHeap)
{
	if (aHeap->size > 0)
	{
		free(aHeap->elt);
	}
}

void heap_insertElt(myHeap *aHeap, double dist, node_type **aNode)
{
	int noOfElt;
	int orgSize;

	heapEntry *elt;

	heapEntry tempEntry;

	int currIndex, parIndex;

	noOfElt = aHeap->noOfElt;

	if (noOfElt == 0)
	{
		elt = aHeap->elt;

		elt[1].aNode = (*aNode);
		elt[1].dist = dist;
		aHeap->noOfElt++;
	}
	else
	{
		// need to allocate?
		orgSize = aHeap->size;
		if (noOfElt+2 >= orgSize)
		{
			aHeap->elt = (heapEntry *) realloc(aHeap->elt, sizeof(heapEntry)*orgSize*2);
			aHeap->size = orgSize*2;
		}

		elt = aHeap->elt;

		// insert
		currIndex = noOfElt+1;
		elt[currIndex].aNode = (*aNode);
		elt[currIndex].dist = dist;

		// parent.dist > child.dist
		parIndex = currIndex/2;
		while ((parIndex >= 1) && (elt[parIndex].dist > elt[currIndex].dist))
		{
			// bubble up
			tempEntry = elt[parIndex];
			elt[parIndex] = elt[currIndex];
			elt[currIndex] = tempEntry;

			currIndex = parIndex;
			parIndex = currIndex/2;
		}

		aHeap->noOfElt++;
	}
}

heapEntry *heap_removeElt(myHeap *aHeap)
{
	heapEntry *returnValue;

	heapEntry *elt;

	double minDist;

	heapEntry tempEntry;

	int noOfElt;

	int currIndex, leftIndex, rightIndex, choseIndex;

	elt = aHeap->elt;

	//
	if (aHeap->noOfElt > 0)
	{
		returnValue = (heapEntry *) malloc(sizeof(heapEntry));
		memcpy(returnValue, &(elt[1]), sizeof(heapEntry));

		//
		currIndex = 1;
		noOfElt = aHeap->noOfElt;
		elt[currIndex] = elt[noOfElt];

		aHeap->noOfElt--;
		noOfElt = aHeap->noOfElt;

		choseIndex = noOfElt+1;


		// find choseIndex
		leftIndex = currIndex*2;
		rightIndex = currIndex*2+1;
		if (rightIndex <= noOfElt)
		{
			if (elt[leftIndex].dist < elt[rightIndex].dist)
			{
				choseIndex = leftIndex;
			}
			else
			{
				choseIndex = rightIndex;
			}
			minDist = elt[choseIndex].dist;
		}
		else
		{
			if (leftIndex <= noOfElt)
			{
				choseIndex = leftIndex;
				minDist = elt[choseIndex].dist;
			}
			else
			{
				// do nothing
			}
		}


		while ((choseIndex <= noOfElt) && (minDist < elt[currIndex].dist))
		{
			// swapping
			tempEntry = elt[choseIndex];
			elt[choseIndex] = elt[currIndex];
			elt[currIndex] = tempEntry;

			currIndex = choseIndex;

			leftIndex = currIndex*2;
			rightIndex = currIndex*2+1;
			if (rightIndex <= noOfElt)
			{
				if (elt[leftIndex].dist < elt[rightIndex].dist)
				{
					choseIndex = leftIndex;
				}
				else
				{
					choseIndex = rightIndex;
				}
				minDist = elt[choseIndex].dist;
			}
			else
			{
				if (leftIndex <= noOfElt)
				{
					choseIndex = leftIndex;
					minDist = elt[choseIndex].dist;
				}
				else
				{
					// do nothing
					break;
				}
			}
			
		}
	}
	else
	{
		returnValue = NULL;
	}
	

	return returnValue;
}
