#ifndef PT_H
#define PT_H

#include <stdio.h>
#include "config.h"
#include <stdlib.h>
#include "set.h"
#include "list.h"
#ifndef WIN32
#include <sys/resource.h>
#include <sys/times.h>
#include <unistd.h>
#endif

/**
 * Struct for a node in preference tree node
 */
typedef struct preferenceTreeNode_s
{
	/** Number of child node */
	int noOfChild;
	/** Array of pointer to child node */
	struct preferenceTreeNode_s **child;
	/**
	 * The dimension of refinement 
	 * -1 indicate no refinement
	 */
	int dimNo;
	/**
	 * The integer representation of A1 in refinement: (A1, others)
	 * -1 indicate empty node
	 */
	int classBefore;
	/** Set of pointID of disqualified point due to condition
	 * in label
	 * Please note that the root node should store the candidate of
	 * skyline points
	 * The must be empty for empty node
	 */
	list *disqualifiedPoint;
	mySet *setDisqualifiedPoint;
} preferenceTreeNode;

preferenceTreeNode* buildPreferenceTree(config *aConfig, myClassSetSetSet* mdc);
preferenceTreeNode* buildPreferenceTreeDescend(config *aConfig, int dimLevel, int classBefore);
//preferenceTreeNode* buildPreferenceTreeDescendOld(config *aConfig, myClassSetSetSet* mdc, int dimLevel, int classBefore, preferenceTreeNode *parent, mySet *Skyline, myClassSet *accumulativePreference);
preferenceTreeNode* buildPreferenceTreeDescendOld(config *aConfig, myClassSetSetSet* mdc, mySet *validIndex, int dimLevel, int classBefore, preferenceTreeNode *parent, mySet *Skyline, myClassSet *accumulativePreference);
void fillPreferenceTreeDescend(preferenceTreeNode *node, int dimLevel, int pointID, int *p);
preferenceTreeNode* preferenceTreeNode_new(int dimNo, int classBefore);
void printPreferenceTree(FILE *filePtr, preferenceTreeNode *node, int *id);
void printPreferenceTreeNode(FILE *filePtr, preferenceTreeNode *node);
//list* preferenceTree_queryDescend(config *aConfig, int dimLevel, preferenceTreeNode *node, list *Skyline, mySetSet *R);
mySet *preferenceTree_queryDescend(config *aConfig, int dimLevel, preferenceTreeNode* node, mySet *gSkyline, mySetSet *R);
//mySet* preferenceTree_skylineQuery(config *aConfig, myClassSetSetSet* mdc);
mySet* preferenceTree_skylineQuery(config *aConfig, myClassSetSetSet* mdc, FILE *debugFilePtr, FILE *timeFilePtr);
void calculateExecutionTime(struct rusage *myTime1, struct rusage *myTime2, float *userTime, float *sysTime);
int calculateSizeOfPreferenceTree(preferenceTreeNode *node);

#ifndef WIN32
extern void calculateExecutionTime(struct rusage *myTime1, struct rusage *myTime2, float *userTime, float *sysTime);
#endif

#endif
